# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from .pslMatrix import *

__all__ = ['BianchiGroupSampler']

class BianchiGroupSampler:
    """
    Produces a list of matrices that from the Bianchi group PSL(2, O) (where O
    is the ring of integers of Q(sqrt(-D))). They will generate PSL(2, O) if
    large enough ranges are given. Such matrices can be obtained by:

    BianchiGroupSampler(ImagQuadraticFieldUtil(Integer(3)), 2, 2).matrices
    
    Here, 2 and 2 indicates the ranges - 2 is large enough to generate the
    Bianchi groups for D = 1, 2, 3, 7, 11, 19.
    """

    def _create_matrix(self, a, b, c, d):
        m = [[1 + a,     b], 
             [    c, 1 - a + d]]

        return PslMatrix(self.imagQuadraticFieldUtil, m)
        
    def __init__(self, imagQuadraticFieldUtil, num_a, num_d):
        self.imagQuadraticFieldUtil = imagQuadraticFieldUtil
                
        self.matrices = []

        # We generate matrices [[1 + a, b], [c, 1 - a + d]]
        # These are all the values for a and d
        values_a = imagQuadraticFieldUtil.enumerate_alg_integers(num_a)
        values_d = imagQuadraticFieldUtil.enumerate_alg_integers(num_d)
        
        values_a.sort(key = abs)
        values_d.sort(key = abs)

        for d in values_d:
            for a in values_a:
                # Compute what the product of b and c must be such that the
                # determinant is 1.
                diag_prod = a*d-a**2+d
                if diag_prod == 0:
                    # If b * c must be 0, try values for b and c from values_a.
                    for bc in values_a:
                        self.matrices.append(self._create_matrix(a,bc,0,d))
                        self.matrices.append(self._create_matrix(a,0,bc,d))
                else:
                    # Otherwise use all possible ways of factoring that number
                    # into b and c
                    for b, c in imagQuadraticFieldUtil.all_products(diag_prod):
                        self.matrices.append(self._create_matrix(a,b,c,d))
