# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

"""
Various helpers around Sage's Polyhedron type.

We assume here that the polyhedron is 3-dimension. The term "face" sometimes
refers to a cell of dimension 2 of a polyhedron and sometimes to a cell of
any dimension as it does in Sage's Polyhedron or Regina.
"""

from sage.all import Polyhedron

from .pslMatrix import *
from .polyhedronUtils import *
from .poincareBall import *

from collections import OrderedDict
  
__all__ = ['DirichletDomain']

def getPolyhedron(ieqs):
    """
    The Sage Polyhedron constructor does not perform well when there are
    millions of inequalities but most will not yield a face of the polyhedron.
    We can get much better performance by slicing all the inequalities into
    batches.
    """

    # Size of a batch to process
    stride = 1000

    # Inequalities of the polyhedron constructed so far
    filtered_ieqs = []

    # For each batch
    for i in range((len(ieqs) + stride - 1) / stride):
        if True:
            print "Processing plane %d/%d (candidate Dirichlet domain has %d faces)" % (
                i * stride, len(ieqs), len(filtered_ieqs))

        # Polyhedron that is intersection of previous polyhedron and new batch
        polyhedron = Polyhedron(
            ieqs = filtered_ieqs + ieqs[i * stride : (i + 1) * stride])
        # Inequalities of this new polyhedron
        filtered_ieqs = list(polyhedron.inequalities())

    return polyhedron

class DirichletDomain:
    """
    Recall that each face of a Dirichlet domain comes with a matrix and a
    mate face such that the inverse of the matrix is taking the face to its
    mate face (or in other words, the face is glued to the mate face of the
    polyhedron translated by the matrix).
    The matrix associate to the mate is the inverse matrix.
    For an (orientation-preserving) involution, a face can be its own mate,
    switching some of the vertices.

    Given a set of generators (which are PslMatrix'es) of a group, try to
    construct a Dirichlet domain.

    The resulting object has the following properties describing the Dirichlet
    domain and the face-pairings:
    * polyhedron which is a Sage Polyhedron representing the Dirichlet domain
      in the Klein model (with one of the axis scaled by sqrt(D)).
    * barycentricSimplices which are BarycentricSimplex'es from subdividing the
      polyhedron
    * vertices_in_poincare_ball which are the coordinates of the vertices of
      the polyhedron in the Poincare model (same order as they are in the
      polyhedron). The type is PoincareBallAlgebraicPoint.
    * vertex_in_poincare_ball_to_index which is a dictionary mapping a point
      to the index where it occurs in vertices_in_poincare_ball.
    * face_to_matrix which contains the face-pairing matrices, i.e., for a tuple
      of indices of vertices forming a face, it contains the PslMatrix pairing
      that face with another face.
    * matrix_to_face which is the inverse map.
    * face_to_opposite_face_and_vertex_to_opposite_vertex. This maps a face
      (again encoded by its vertex indices) to a pair consisting of the
      mate face as well as another map indicating how the vertices correspond
      to the vertices of the other face under the pairing (the map contains
      the index of a vertex).
    """

    @staticmethod
    def _compute_plane_eqn_to_matrix(generators):
        """
        Returns a dictionary containing for each non-identity PslMatrix in
        generators the equation for the plane (one entry being scaled
        by sqrt(D) so that it can have integral coefficients) that is bisecting
        the origin and the image of the origin by the given PslMatrix in the
        Klein model.
        """
        # Identity matrix
        identity = PslMatrix.identity(generators.imagQuadraticFieldUtil)

        # Origin in Poincare ball model
        origin = PoincareBallQuadraticFieldPoint.origin(
            generators.imagQuadraticFieldUtil)
        
        def plane_eqn(matrix):
            """
            Let matrix act on origin and compute bisecting plane.
            """
            try:
                return matrix.act(
                    origin).toScaledRationalBisectingPlaneEquationInKleinModel()
            except Exception as e:
                print "matrix: %r" % matrix
                raise e
        
        return OrderedDict([ 
                (plane_eqn(matrix), matrix)
                for matrix in generators.matrices
                if not matrix == identity])

    def _computePolyhedron(self, generators):
        """
        Compute the polyhedron forming the Dirichlet domain using the
        PslMatrix'es in generators.

        Also compute face_to_matrix.
        """

        # Remember the correspondence of planes (or more precisely their
        # defining equations) to PslMatrix'es.
        plane_eqn_to_matrix = DirichletDomain._compute_plane_eqn_to_matrix(
            generators)

        # Compute polyhedron as intersection of half spaces
        # self.polyhedron = Polyhedron(ieqs = plane_eqn_to_matrix.keys())
        self.polyhedron = getPolyhedron(ieqs = plane_eqn_to_matrix.keys())

        # Get all faces of Sage polyhedron and compute face_to_matrix
        self.face_to_matrix = {
            polyhedron_face_to_vertex_indices(polyhedron_face) :
                plane_eqn_to_matrix[
                    polyhedron_face_to_plane_eqn(polyhedron_face)]
            for polyhedron_face in self.polyhedron.faces(2) }

    def __init__(self, generators):
        self.imagQuadraticFieldUtil = generators.imagQuadraticFieldUtil

        # Compute polyhedron as intersection of half spaces
        self._computePolyhedron(generators)

        # Compute barycentric subdivision
        self.barycentricSimplices = polyhedron_to_barycentric_subdivision(
            self.polyhedron)

        self.matrix_to_face = {
            matrix : face for face, matrix in self.face_to_matrix.items() }
            
        self.vertices_in_poincare_ball = [
            PoincareBallAlgebraicPoint.fromScaledRationalPointInKleinModel(
                self.imagQuadraticFieldUtil, v.vector())
            for v in self.polyhedron.vertices() ]
            
        self.vertex_in_poincare_ball_to_index = {
            v : i for i, v in enumerate(self.vertices_in_poincare_ball) }

        def opposite_face_and_vertex_to_opposite_vertex(face, matrix):
            """
            Takes a face and the matrix pairing it with another face.
            Returns a pair of the mate face and a mapping of the indices
            of the vertices of the face to that of the mate face.
            """

            inv = matrix.inverse()

            def opposite_vertex(v):
                """
                Takes a vertex. Gives the correponding vertex under face-pairing.
                """

                try:
                    return self.vertex_in_poincare_ball_to_index[
                        inv.act(self.vertices_in_poincare_ball[v])] 
                except KeyError as e:
                    raise Exception(
                        "When computing Dirichlet domain, could not match vertex of face to mate face: %s" % e.args[0])

            try:
                return (
                    self.matrix_to_face[inv],
                    { v : opposite_vertex(v) for v in face } )
            except KeyError as e:
                raise Exception(
                    "When computing Dirichlet domain, could not find mate face corresponding to %s" % e.args[0])

        self.face_to_opposite_face_and_vertex_to_opposite_vertex = {
            face : opposite_face_and_vertex_to_opposite_vertex(face,matrix)
            for face, matrix in self.face_to_matrix.items() }
                    
