# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from .imagQuadraticFieldUtil import *

from sage.all import QQ, vector, Matrix, frac

__all__ = ['Ideal']

def _move_point_to_fundamental_domain(m, minv, p):
    """
    Reduces p with respect to the lattice described by the
    2x2 rational matrix m.
    """
    return (p * minv).apply_map(frac, QQ) * m

def _reduce_step(a,b,c):
    """
    Reduces c with respect to the lattice spanned by a and b.
    """
    m = Matrix([a,b])
    if m.det() == 0:
        # If a and b are linearly dependent, do nothing.
        return a, b, c
    new_c = _move_point_to_fundamental_domain(m, m.inverse(), c)
    return a, b, new_c

def _reduce_triple(a,b,c):
    """
    Keep reducing one vector by the other two vector until zero.
    Return the two non-zero vectors.
    """

    while True:
        if a == vector([0,0]):
            return b, c
        if b == vector([0,0]):
            return a, c
        if c == vector([0,0]):
            return a, b
        c, a, b = _reduce_step(a, b, c)

def _reduce_vectors(v):
    """
    Given a set of 2-vectors v, return two vectors spanning the same
    lattice.

    The first two vectors of v are assumed to be linearly independent.
    """

    v0, v1 = v[:2]
    for v2 in v[2:]:
        v0, v1 = _reduce_triple(v0, v1, v2)
    return v0, v1

class Ideal:
    """
    Represents an ideal I in the ring of integers O in Q(sqrt(-D)).

    The main purpose is the modulo method.
    """

    def __init__(self, imagQuadraticFieldUtil, gens):
        """
        Constructs an ideal given an instance imagQuadraticFieldUtil of
        ImagQuadraticFieldUtil and generators gens that are elements in
        imagQuadraticFieldUtil.CNF (Sage's QuadraticField(-D)).
        """

        # The generators
        self.gens = gens
        self._imagQuadraticFieldUtil = imagQuadraticFieldUtil
        self.D = imagQuadraticFieldUtil.D

        # Basis for the lattice formed by the ring of integers
        basis = [ self._imagQuadraticFieldUtil.CNF(1),
                  self._imagQuadraticFieldUtil.algInt ]

        # Vectors spanning the lattice formed by this ideal
        lattice_vecs = [
            vector(QQ, ImagQuadraticFieldUtil.getCoefficientsOfElement(gen * b))
            for gen in gens for b in basis ]

        # Construct a basis of two linearly independeny vectors for the lattice
        # formed by this ideal
        lattice_basis = _reduce_vectors(lattice_vecs)

        # Matrices to take that lattice to the lattice Z + Z
        self._matrix = Matrix(lattice_basis)
        self._inv_matrix = self._matrix.inverse()

        # Cache for acceleration
        self._modulo_cache = {}

    def _modulo(self, z):
        # Convert z to a vector
        pt = vector(QQ, ImagQuadraticFieldUtil.getCoefficientsOfElement(z))
        # Reduce vector using lattice
        pt = _move_point_to_fundamental_domain(
            self._matrix, self._inv_matrix, pt)
        # Convert back to an elment in QuadraticField(-D)
        return self._imagQuadraticFieldUtil.getAlgebraicInteger(*pt)
         
    def modulo(self, z):
        """
        Computes a canonical representative for an element in O/I given any
        representative in O.
        """
        if not self._modulo_cache.has_key(z):
            self._modulo_cache[z] = self._modulo(z)

        return self._modulo_cache[z]
