# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

"""
Various helpers around Sage's Polyhedron type.

We assume here that the polyhedron is 3-dimension. The term "face" sometimes
refers to a cell of dimension 2 of a polyhedron and sometimes to a cell of
any dimension as it does in Sage's Polyhedron or Regina.
"""

from sage.all import lcm, gcd

__all__ = ['polyhedron_face_to_vertex_indices',
           'polyhedron_face_to_plane_eqn',
           'BarycentricSimplex',
           'polyhedron_to_barycentric_subdivision',
           'make_plane_equation_integral']

def polyhedron_face_to_vertex_indices(face):
    """
    Given a face of a Sage Polyhedron (of any dimension), return all the
    indices of the vertices forming that face.
    """
    
    return tuple([v.index() for v in face.vertices()])

def polyhedron_face_to_plane_eqn(face):
    """
    Given a 2-dimension of a 3-dimensional polyhedron, return the equation
    of the plane it lies in.
    """
    return tuple(face.ambient_Hrepresentation()[0].vector())

class BarycentricSimplex(tuple):
    """
    We capture the combinatorics of a Sage Polyhedron by barycentrically
    subdividing it.

    The simplices of the barycentric subdivision are parametrized by a
    tuple (face, edge, vertex) such that the simplex has a vertex at the
    center of the polyhedron, center of the face, center of the edge, and at
    the given vertex.
    The face in the tuple is actually just a tuple of the indices of vertices
    forming that face. Similar for edge. The last entry of (face, edge, vertex)
    is just the index of the vertex in the polyhedron.
    """
    pass

def _boundary_map(face):
    """
    Given a face of a Sage Polyhedron of dimension k, return all the adjacent
    faces of dimension k - 1.
    """

    # The adjacency relation is actually encoded as face lattice poset.
    return face.polyhedron().face_lattice().lower_covers(face)

def polyhedron_to_barycentric_subdivision(polyhedron):
    """
    Give the barycentric subdivision of a Sage Polyhedron as a list of
    BarycentricSimplex'es.
    """

    return [
        BarycentricSimplex(
            (polyhedron_face_to_vertex_indices(face),
             polyhedron_face_to_vertex_indices(edge),
             polyhedron_face_to_vertex_indices(vert)[0]))
        for face in polyhedron.faces(2)
        for edge in _boundary_map(face)
        for vert in _boundary_map(edge) ]

def _eliminate_denominators(v):
    common_denominator = lcm([x.denom() for x in v])
    return [ x * common_denominator for x in v]

def _eliminate_common_divisor(v):
    g = gcd(v)
    return [ x / g for x in v]

def make_plane_equation_integral(v):
    """
    Given a vector of Sage Rational's, multiply by a number such all entries
    become integral without a common divisior.
    """
    return tuple(
        _eliminate_common_divisor(
            _eliminate_denominators(v)))
