import regina

__all__ = ['barycentricCoarsening']

def barycentricCoarsening(T):
    """
    T is assumed to be a regina Triangulation coming from the
    barycentric subdivision of a polyhedral cell complex where the
    vertices of the simplices in the subdivision are labeled such that
    0 corresponds to a vertex of a polyhedron, 1 to the center of an
    edge of a polyhedron, 2 to the center of a face of polyhedron, and
    3 to the center of a polyhedron.

    This function produces a new regina triangulation with a fourth
    of the number of simplices.

    Namely, a new simplex will have its vertex 0 and 1 at the ends
    of an edge of a polyhedron and vertex 2 and 3 at the centers of
    two adjacent polyhedra.

    In other words, a new simplex is the union of four old simplices
    glued together along faces 0 and 3.
    """

    pId = regina.Perm4()
    p01 = regina.Perm4(1,0,2,3)
    p23 = regina.Perm4(0,1,3,2)

    result = regina.Triangulation3()

    # A list with an entry per old tetrahedra. Each entry is a pair
    # (new Tetrahedron, perm) where perm describes how the vertices of
    # the new tetrahedron correspond to the one of the old
    # tetrahedron.
    # Note that for each new Tetrahedron, there will be four
    # tuples (new Tetrahedron, perm) in total.
    oldToNew = T.size() * [ None ]

    for i in range(T.size()):
        # Only process an old tetrahderon once
        if oldToNew[i]:
            continue

        newTet = result.newSimplex()
        tet = T.simplex(i)
        # The new simplex corresponds to the four old simplices
        # obtained by going across face 0 and 3
        for p0, tet0 in [(pId, tet), (p01, tet.adjacentSimplex(0))]:
            for p3, tet3 in [(pId, tet0), (p23, tet0.adjacentSimplex(3))]:
                oldToNew[tet3.index()] = (newTet, p0 * p3)

    # Construct new triangulation using dict, looking at the old
    # tetrahedra to glue the new ones.
    for i in range(T.size()):
        tet = T.simplex(i)
        newTet, p = oldToNew[i]

        # Faces 0 and 3 of the old tetrahedron are "internal" to the
        # new tetrahedron.
        # But when two old tetrahedra are glued along face 1 or 2, we
        # need to glue together the corresponding new tetrahedra
        # using the right permutation.

        for n in [1, 2]:
            otherTet = tet.adjacentSimplex(n).index()
            otherNewTet, otherP = oldToNew[otherTet]

            # Luckily, we do not need to think in what order the
            # permutations or their inverses should be multiplied,
            # since all permutations in oldToNew are involutions and
            # commute.
            newTet.join(p[n], otherNewTet, p * otherP)

    return result
